import asyncio
import enum
import functools
import inspect
import json
import typing

from miio.click_common import DeviceGroupMeta

from .. import fhem, generic, utils


class miio(generic.FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self._set_list = {}
        self._device = None
        self._fct_update_tasks = {}
        self._cmd_lock = asyncio.Lock()
        self._attr_update_functions = ""
        self._attr_list = {
            "update_functions": {
                "default": "status:60,info:600",
                "help": (
                    "Define command which should be executed every X seconds.<br>"
                    "info:600 executes info every 600s<br>Default: status:60,info:600"
                ),
            }
        }
        self.set_attr_config(self._attr_list)
        return

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        self.hash = hash
        if len(args) < 6:
            return "Usage: define miiodev fhempy miio <TYPE> <IP> <TOKEN>"
        self._miio_devtype = args[3]
        self._miio_ip = args[4]
        self._miio_token = args[5]

        self._miio_device_class = None
        for device_class in DeviceGroupMeta._device_classes:
            if device_class.get_device_group().name == self._miio_devtype:
                self._miio_device_class = device_class
                break
        if self._miio_device_class is None:
            return f"Device {self._miio_devtype} not found."

        for dev_cmd in self._miio_device_class.get_device_group().commands.keys():
            self._set_list[dev_cmd] = {"function": "set_command", "default": None}
            fct = self._miio_device_class._device_group_commands[dev_cmd].func
            sig = inspect.signature(fct)
            self._set_list[dev_cmd]["help"] = []
            if len(list(sig.parameters)) > 1:
                self._set_list[dev_cmd]["args"] = []
                for par in sig.parameters:
                    if sig.parameters[par].name == "self":
                        continue
                    self._set_list[dev_cmd]["args"].append(sig.parameters[par].name)
                    self._set_list[dev_cmd]["help"].append(str(sig.parameters[par]))
                    if len(list(sig.parameters)) == 2:
                        # set options if there is only one parameter
                        annot = sig.parameters[par].annotation
                        if not inspect.isclass(annot):
                            self.logger.error("Annotation is not class: " + str(annot))
                        if inspect.isclass(annot) and issubclass(annot, enum.Enum):
                            self._set_list[dev_cmd]["options"] = ",".join(
                                list(map(lambda x: x.name, annot))
                            )
                        elif inspect.isclass(annot) and issubclass(annot, bool):
                            self._set_list[dev_cmd]["options"] = "on,off"

            helptext = ""
            if (
                "help" in self._miio_device_class._device_group_commands[dev_cmd].kwargs
                and self._miio_device_class._device_group_commands[dev_cmd].kwargs[
                    "help"
                ]
                is not None
            ):
                helptext = self._miio_device_class._device_group_commands[
                    dev_cmd
                ].kwargs["help"]
            self._set_list[dev_cmd]["help"] = (
                helptext + "<br>Arguments: " + " ".join(self._set_list[dev_cmd]["help"])
            )

        self.set_set_config(self._set_list)
        self._device = self._miio_device_class(ip=self._miio_ip, token=self._miio_token)
        await fhem.readingsSingleUpdateIfChanged(hash, "state", "active", 1)
        await self.set_attr_update_functions(self.hash)

    async def set_attr_update_functions(self, hash):
        for task in self._fct_update_tasks.copy():
            self._fct_update_tasks[task].cancel()
            del self._fct_update_tasks[task]

        if self._attr_update_functions != "":
            fct_upd_list = self._attr_update_functions.split(",")
            for fct_upd in fct_upd_list:
                sec = int(fct_upd.split(":")[1])
                fct = fct_upd.split(":")[0]
                self._fct_update_tasks[fct] = self.create_async_task(
                    self.fct_update_loop(fct, sec)
                )

    async def fct_update_loop(self, fct_name, sec):
        while True:
            try:
                async with self._cmd_lock:
                    await asyncio.sleep(1)
                    await self.send_command(fct_name, None, raise_exc=True)
                await fhem.readingsSingleUpdateIfChanged(
                    self.hash, "presence", "online", 1
                )
            except Exception:
                if fct_name != "status":
                    self.logger.error(f"Failed to send_command: {fct_name}")
                await fhem.readingsSingleUpdateIfChanged(
                    self.hash, "presence", "offline", 1
                )
            await asyncio.sleep(sec)

    async def set_command(self, hash, params):
        cmd = params["cmd"]
        self.create_async_task(self.send_command(cmd, params))

    def is_number(self, string):
        try:
            float(string)
            return True
        except ValueError:
            return False

    async def send_command(self, fct_name, params, raise_exc=False):
        fct = self._device._device_group_commands[fct_name].func
        sig = inspect.signature(fct)
        args = []
        for par_name in sig.parameters:
            if par_name == "self":
                continue
            if par_name not in params or params[par_name] is None:
                continue
            ann = sig.parameters[par_name].annotation
            if inspect.isclass(ann) and issubclass(ann, enum.Enum):
                args.append(ann[params[par_name]])
            elif inspect.isclass(ann) and issubclass(ann, bool):
                args.append(params[par_name] == "on")
            elif ann == typing.List:
                args.append(json.loads(params[par_name]))
            elif hasattr(ann, "_name") and ann._name == "Tuple":
                args.append(tuple(map(int, params[par_name].split(","))))
            elif ann == inspect.Signature.empty:
                if self.is_number(params[par_name]):
                    args.append(float(params[par_name]))
                else:
                    args.append(params[par_name])
            else:
                args.append(ann(params[par_name]))
        # call function with arguments
        call_fct = getattr(self._device, fct_name)
        try:
            reply = await utils.run_blocking(functools.partial(call_fct, *args))
        except Exception as exc:
            if raise_exc:
                raise exc
            self.logger.exception(f"Failed to call {fct_name}")
            return

        # handle reply
        if hasattr(reply, "__dict__"):
            await fhem.readingsBeginUpdate(self.hash)
            try:
                st = dict(
                    (x, getattr(reply, x))
                    for x in reply.__class__.__dict__
                    if isinstance(reply.__class__.__dict__[x], property)
                )
                for prop in st:
                    if prop == "raw":
                        continue
                    try:
                        for data_name in st[prop]:
                            await fhem.readingsBulkUpdate(
                                self.hash,
                                prop + "_" + data_name,
                                st[prop][data_name],
                            )
                    except Exception:
                        await fhem.readingsBulkUpdate(self.hash, prop, st[prop])
            except Exception:
                self.logger.exception("Failed to update readings")
            await fhem.readingsEndUpdate(self.hash, 1)
        else:
            if (
                reply is not None
                and isinstance(reply, str)
                and reply.lower() != "['ok']"
            ):
                await fhem.readingsSingleUpdate(self.hash, fct.__name__, reply, 1)
        if fct_name != "status":
            # wait before status request as it takes some time for the miio device
            # to change the state
            await asyncio.sleep(2)
            await self.send_command("status", None)



import asyncio
import functools
import json

import aiohttp
from bs4 import BeautifulSoup
from fhempy.lib import utils

from .. import fhem, generic


class google_weather(generic.FhemModule):
    def __init__(self, logger):
        super().__init__(logger)

        attr_config = {
            "interval": {
                "default": 61,
                "format": "int",
                "help": "Change interval in minutes, default is 61.",
            }
        }
        self.set_attr_config(attr_config)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        if len(args) != 4:
            return "Usage: define my_weather fhempy google_weather CITY"

        self.city = args[3]
        self.update_url = f"https://www.google.com/search?q={self.city}+weather"

        self.create_async_task(self.update_loop())

    async def update_loop(self):
        headers = {
            "accept": (
                "text/html,application/xhtml+xml,application/xml;"
                "q=0.9,image/avif,image/webp,image/apng,*/*;"
                "q=0.8,application/signed-exchange;v=b3;q=0.9"
            ),
            "accept-language": (
                "en-DE,en;q=0.9,de-DE;q=0.8,de;q=0.7,en-GB;q=0.6,en-US;q=0.5"
            ),
            "user-agent": (
                "Mozilla/5.0 (X11; CrOS x86_64 14909.100.0) AppleWebKit/537.36"
                " (KHTML, like Gecko) Chrome/104.0.0.0 Safari/537.36"
            ),
            "sec-ch-ua": (
                '"Chromium";v="104", " Not A;Brand";v="99", "Google Chrome";v="104"'
            ),
            "sec-ch-ua-arch": '"x86"',
            "sec-ch-ua-bitness": '"64"',
            "sec-ch-ua-full-version": '"104.0.5112.83"',
            "sec-ch-ua-full-version-list": (
                '"Chromium";v="104.0.5112.83", " Not A;Brand";v="99.0.0.0", '
                '"Google Chrome";v="104.0.5112.83"'
            ),
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-model": '""',
            "sec-ch-ua-platform": '"Chrome OS"',
            "sec-ch-ua-platform-version": '"14909.100.0"',
            "sec-ch-ua-wow64": "?0",
            "sec-fetch-dest": "document",
            "sec-fetch-mode": "navigate",
            "sec-fetch-site": "none",
            "sec-fetch-user": "?1",
            "upgrade-insecure-requests": "1",
            "x-chrome-connected": (
                "source=Chrome,mode=0,enable_account_consistency=true,"
                "supervised=false,consistency_enabled_by_default=false"
            ),
        }
        while True:
            # aiohttp get
            try:
                async with aiohttp.ClientSession(trust_env=True) as session:
                    async with session.get(self.update_url, headers=headers) as resp:
                        if resp.status == 200:
                            await self.handle_response(await resp.text())
                        else:
                            await fhem.readingsSingleUpdate(
                                self.hash,
                                "state",
                                f"failed: HTTP error {resp.status}",
                                1,
                            )
                            self.logger.error(
                                f"Failed to fetch {self.update_url}, "
                                f"failed with status {resp.status}"
                            )
            except Exception:
                self.logger.exception("Failed to update")
            await asyncio.sleep(self._attr_interval * 60)

    def soup_extract(self, soup, element, search_obj):
        element = soup.find(element, search_obj)
        if element is None:
            self.logger.error(f"Element {element} with {search_obj} not found!")
        return element

    def soup_extract_text(self, soup, element, search_obj):
        element = self.soup_extract(soup, element, search_obj)
        if element is None:
            return "-"
        return element.text

    def get_current_temperature(self, soup):
        return self.soup_extract_text(soup, "span", {"id": "wob_tm"})

    def get_current_condition(self, soup):
        return self.soup_extract(soup, "img", {"wob_tci"})

    def get_current_windspeed(self, soup):
        return self.soup_extract_text(soup, "span", {"id": "wob_ws"})

    def get_last_update(self, soup):
        return self.soup_extract_text(soup, "div", {"id": "wob_dts"})

    def get_location_name(self, soup):
        return self.soup_extract_text(soup, "div", {"id": "wob_loc"})

    def get_current_precipitation(self, soup):
        return self.soup_extract_text(soup, "span", {"id": "wob_pp"})

    def get_current_humidity(self, soup):
        return self.soup_extract_text(soup, "span", {"id": "wob_hm"}).replace("%", "")

    def get_next_days(self, soup):
        next_days = []
        days = soup.find("div", attrs={"id": "wob_dp"})
        for day in days.findAll("div", attrs={"class": "wob_df"}):
            # extract the name of the day
            day_name = day.findAll("div")[0].attrs["aria-label"]
            # get weather status for that day
            image = day.find("img")
            weather = image.attrs["alt"]
            temp = day.findAll("span", {"class": "wob_t"})
            # maximum temparature in Celsius, use temp[1].text if you want fahrenheit
            max_temp = temp[0].text
            # minimum temparature in Celsius, use temp[3].text if you want fahrenheit
            min_temp = temp[2].text
            next_days.append(
                {
                    "name": day_name,
                    "weather": weather,
                    "max_temp": max_temp,
                    "min_temp": min_temp,
                    "image": image,
                }
            )
        return next_days

    def extract_weather(self, response):
        soup = BeautifulSoup(response, "html.parser")
        self.cur_temp = self.get_current_temperature(soup)
        cur_condition_element = self.get_current_condition(soup)
        if cur_condition_element is None:
            self.cur_condition = "-"
            self.cur_condition_img = "-"
        else:
            self.cur_condition = cur_condition_element["alt"]
            self.cur_condition_img = f"{cur_condition_element}"
        self.cur_windspeed = self.get_current_windspeed(soup)
        self.cur_windspeed_number = self.cur_windspeed.split(" ")[0]
        self.cur_precipitation = self.get_current_precipitation(soup).replace("%", "")
        self.cur_humidity = self.get_current_humidity(soup)
        self.next_days = self.get_next_days(soup)
        self.last_update = self.get_last_update(soup)
        self.location = self.get_location_name(soup)

        wdata = None
        for script in soup.find_all("script"):
            if script.text.find("\\x22wobnm\\x22") > -1:
                wdata = script.text
                break
        if wdata is not None:
            start_json = wdata.find("\\x22wobnm\\x22")
            end_json = wdata.find("';google.pmc=")
            weather_json = json.loads(
                "{" + wdata[start_json:end_json].replace("\\x22", '"')
            )

        self.next_hours = []
        for x in weather_json["wobnm"]["wobhl"]:
            self.next_hours.append(x)
            if len(self.next_hours) > 24:
                break

    async def handle_response(self, response):
        # bs4
        await utils.run_blocking(functools.partial(self.extract_weather, response))

        await fhem.readingsBeginUpdate(self.hash)
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "cur_temperature", self.cur_temp
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "cur_humidity", self.cur_humidity
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "cur_precipitation", self.cur_precipitation
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "cur_weather", self.cur_condition
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "cur_weather_img", f"<html>{self.cur_condition_img}</html>"
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "cur_windspeed_with_unit", self.cur_windspeed
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "cur_windspeed", self.cur_windspeed_number
        )

        i = 0
        for day in self.next_days:
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, f"next_days_{i}_name", day["name"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, f"next_days_{i}_weather", day["weather"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, f"next_days_{i}_weather_img", f"<html>{day['image']}</html>"
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, f"next_days_{i}_max_temp", day["max_temp"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, f"next_days_{i}_min_temp", day["min_temp"]
            )
            i += 1

        i = 0
        for nh in self.next_hours:
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, f"next_hours_{i:02d}_condition", nh["c"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, f"next_hours_{i:02d}_time", nh["dts"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, f"next_hours_{i:02d}_humidity", nh["h"].replace("%", "")
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash,
                f"next_hours_{i:02d}_img",
                '<html><img src="' + nh["iu"] + '"/></html>',
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, f"next_hours_{i:02d}_precipitation", nh["p"].replace("%", "")
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, f"next_hours_{i:02d}_temperature", nh["tm"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash,
                f"next_hours_{i:02d}_windspeed",
                nh["ws"].replace(" km/h", ""),
            )
            i += 1

        await fhem.readingsBulkUpdateIfChanged(self.hash, "location", self.location)
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "last_update", self.last_update
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash,
            "state",
            (
                f"<html>{self.cur_condition_img}<br>{self.cur_temp}°C / "
                f"{self.cur_precipitation}% / {self.cur_windspeed}</html>"
            ),
        )
        await fhem.readingsEndUpdate(self.hash, 1)



import asyncio

import aiohttp
from bs4 import BeautifulSoup
from fhempy.lib import utils

from .. import fhem, generic


class ddnssde(generic.FhemModule):

    headers = {
        "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,"
        + "image/avif,image/webp,image/apng,*/*;"
        + "q=0.8,application/signed-exchange;v=b3;q=0.9",
        "Accept-Encoding": "gzip, deflate, br",
        "Accept-Language": "en-AT,en;q=0.9,de-AT;q=0.8,de;"
        + "q=0.7,en-GB;q=0.6,en-US;q=0.5",
        "Cache-Control": "max-age=0",
        "Connection": "keep-alive",
        "Content-Length": "58",
        "Content-Type": "application/x-www-form-urlencoded",
        "Host": "ddnss.de",
        "Origin": "https://ddnss.de",
        "Referer": "https://ddnss.de/login.php",
        "Sec-Fetch-Dest": "document",
        "Sec-Fetch-Mode": "navigate",
        "Sec-Fetch-Site": "same-origin",
        "Sec-Fetch-User": "?1",
        "Upgrade-Insecure-Requests": "1",
        "User-Agent": "Mozilla/5.0 (X11; CrOS x86_64 14989.36.0) "
        + "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/105.0.0.0 Safari/537.36",
        "sec-ch-ua": '"Google Chrome";v="105", "Not)A;Brand";v="8", "Chromium";v="105"',
        "sec-ch-ua-mobile": "?0",
        "sec-ch-ua-platform": "Chrome OS",
    }

    def __init__(self, logger):
        super().__init__(logger)

        self.ip_infos = {}
        self.update_key_ready = asyncio.Event()
        self.hostname_ready = asyncio.Event()

        attr_config = {
            "ip_check_interval": {
                "default": 15,
                "format": "int",
                "help": "Change interval in minutes, default is 15.",
            },
            "hostname": {
                "default": "-",
                "options": "all",
                "help": "Select the hostname you would like to update",
            },
        }
        self.set_attr_config(attr_config)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        if len(args) != 5:
            return "Usage: define ddnss.updater fhempy ddnssde USERNAME PASSWORD"

        self.username = args[3]
        self.password = args[4]

        self.current_ip = await fhem.ReadingsVal(self.hash["NAME"], "current_ip", "")

        self.create_async_task(self.update_ddnss_data_loop())
        self.create_async_task(self.update_ip_loop())

    async def update_readings(self, readings):
        await fhem.readingsBeginUpdate(self.hash)
        for name in readings:
            await fhem.readingsBulkUpdateIfChanged(self.hash, name, readings[name])
        await fhem.readingsEndUpdate(self.hash, 1)

    async def get_json_data(self, url):
        async with aiohttp.ClientSession(trust_env=True) as session:
            async with session.get(url, headers=ddnssde.headers) as resp:
                if resp.status == 200:
                    return await resp.json()
        return None

    async def get_current_ip(self):
        return await self.get_json_data("https://wtfismyip.com/json")

    async def check_myip(self):
        # sorry for the wording, it comes from WTFismyip service
        ip_infos = await self.get_current_ip()
        self.ip_infos = {
            "current_ip": ip_infos["YourFuckingIPAddress"],
            "isp_location": ip_infos["YourFuckingLocation"],
            "isp_hostname": ip_infos["YourFuckingHostname"],
            "isp": ip_infos["YourFuckingISP"],
            "isp_country": ip_infos["YourFuckingCountryCode"],
        }
        await self.update_readings(self.ip_infos)
        return self.ip_infos

    async def is_ip_updated(self):
        ip_infos = await self.check_myip()
        new_ip = ip_infos["current_ip"]
        if new_ip != self.current_ip:
            self.current_ip = new_ip
            return True
        return False

    async def update_ddnss_data(self):
        await self.login()
        ret = await self.retrieve_ddnss_data()
        if ret is False:
            self.logger.error("Failed to retrieve ddnss data")
            return
        await self.update_ddnss_readings()

    async def update_ddnss_data_loop(self):
        while True:
            await self.update_ddnss_data()
            await asyncio.sleep(7500)

    async def update_ddnss_readings(self):
        readings = {
            "update_key": self.update_key,
            "mail": self.ddnss_mail,
            "last_login": self.ddnss_login,
        }

        hostattr = ["all"]
        for host in self.hostnames:
            host_underscore = host.replace(".", "_")
            hostattr.append(host_underscore)
            if host_underscore == self._attr_hostname:
                readings["state"] = host + ": " + self.hostnames[host]["ip"]
            readings[f"host_{host_underscore}"] = host
            readings[f"host_{host_underscore}_ip"] = self.hostnames[host]["ip"]
            readings[f"host_{host_underscore}_last_upd"] = self.hostnames[host][
                "last_update"
            ]
        await self.update_readings(readings)

        # update attribute hostname
        self._conf_attr["hostname"]["options"] = ",".join(hostattr)
        self.set_attr_config(self._conf_attr)
        await utils.handle_define_attr(self._conf_attr, self, self.hash)

    async def login(self):
        async with aiohttp.ClientSession(trust_env=True) as session:
            async with session.get("https://ddnss.de/") as resp:
                if resp.status != 200:
                    raise Exception(f"Open ddnss.de failed with status {resp.status}")
            async with session.get("https://ddnss.de/login.php") as resp:
                if resp.status != 200:
                    raise Exception(f"Open ddnss.de failed with status {resp.status}")

            async with session.post(
                "https://ddnss.de/do.php",
                data={
                    "action": "login",
                    "username": self.username,
                    "passwd": self.password,
                },
                headers=ddnssde.headers,
            ) as resp:
                if resp.status != 200:
                    raise Exception(
                        f"Failed to login to ddnss.de, HTTP status {resp.status}"
                    )

            self.cookie_jar = session.cookie_jar

    async def retrieve_ddnss_data(self):
        async with aiohttp.ClientSession(cookie_jar=self.cookie_jar) as session:
            async with session.get(
                "https://ddnss.de/ua/index.php", headers=ddnssde.headers
            ) as resp:
                data = await resp.text()
                soup = BeautifulSoup(data, "html.parser")
                well_class = soup.find("div", {"class": "well"})
                if well_class is None:
                    return False
                self.update_key = well_class.find("b").text
                self.ddnss_mail = well_class.findAll("p")[1].text
                self.ddnss_login = well_class.findAll("p")[2].text
                self.update_key_ready.set()

            async with session.get(
                "https://ddnss.de/ua/vhosts_list.php", headers=ddnssde.headers
            ) as resp:
                data = await resp.text()
                soup = BeautifulSoup(data, "html.parser")
                self.hostnames = {}
                hostarr = soup.find("tbody").findAll("tr")
                for host in hostarr:
                    hostdetailsarr = host.findAll("td")
                    hostname = hostdetailsarr[1].find("u").text.split()[0]
                    ip = hostdetailsarr[2].text.split()[-1]
                    last_upd = hostdetailsarr[3].text
                    self.hostnames[hostname] = {"ip": ip, "last_update": last_upd}

            return True

    async def update_ip_loop(self):
        while True:
            if self._attr_hostname == "-":
                await fhem.readingsSingleUpdate(
                    self.hash, "state", "Please set hostname attribute", 1
                )
                await self.hostname_ready.wait()

            await self.update_key_ready.wait()
            if await self.is_ip_updated():
                await self.set_ddnss_ip()
            await asyncio.sleep(self._attr_ip_check_interval * 60)

    async def set_attr_hostname(self, hash):
        if self._attr_hostname != "-":
            self.hostname_ready.set()
            self.hostname_ready.clear()
        else:
            await fhem.readingsSingleUpdate(
                self.hash, "state", "Please set hostname attribute", 1
            )

    async def set_ddnss_ip(self):
        async with aiohttp.ClientSession(cookie_jar=self.cookie_jar) as session:
            async with session.get(
                f"http://ddnss.de/upd.php?key={self.update_key}"
                + f"&host={self._attr_hostname.replace('_','.')}",
                headers=ddnssde.headers,
            ) as resp:
                if resp.status == 200:
                    await self.update_ddnss_data()
                else:
                    await fhem.readingsSingleUpdateIfChanged(
                        self.hash, "state", "failed to update", 1
                    )



import asyncio

from greeclimate.device import (
    Device,
    FanSpeed,
    HorizontalSwing,
    Mode,
    TemperatureUnits,
    VerticalSwing,
)
from greeclimate.discovery import Discovery

from .. import fhem, generic


class gree_climate(generic.FhemModule):
    def __init__(self, logger):
        super().__init__(logger)

        attr_config = {
            "interval": {
                "default": 60,
                "format": "int",
                "help": "Change interval, default is 60.",
            }
        }
        self.set_attr_config(attr_config)

        set_config = {
            "mode": {
                "args": ["mode"],
                "argsh": ["mode"],
                "params": {"mode": {"default": "Auto"}},
                "options": "Auto,Cool,Dry,Fan,Heat",
            },
            "desiredTemp": {
                "args": ["temperature"],
                "params": {"temperature": {"format": "int"}},
                "options": "slider,8,1,30",
            },
            "fan_speed": {
                "args": ["speed"],
                "options": "Auto,Low,MediumLow,Medium,MediumHigh,High",
            },
            "swing_horizontal": {
                "args": ["mode"],
                "options": (
                    "Default,FullSwing,Left,LeftCenter,Center,RightCenter,Right"
                ),
            },
            "swing_vertical": {
                "args": ["mode"],
                "options": (
                    "Default,FullSwing,FixedUpper,FixedUpperMiddle,FixedMiddle,"
                    "FixedLowerMiddle,FixedLower,SwingUpper,SwingUpperMiddle,"
                    "SwingMiddle,SwingLowerMiddle,SwingLower"
                ),
            },
            "fresh_air": {"args": ["onoff"], "options": "on,off"},
            "xfan": {"args": ["onoff"], "options": "on,off"},
            "anion": {"args": ["onoff"], "options": "on,off"},
            "sleep": {"args": ["onoff"], "options": "on,off"},
            "light": {"args": ["onoff"], "options": "on,off"},
            "quiet": {"args": ["onoff"], "options": "on,off"},
            "turbo": {"args": ["onoff"], "options": "on,off"},
            "steady_heat": {"args": ["onoff"], "options": "on,off"},
            "power_save": {"args": ["onoff"], "options": "on,off"},
            "on": {},
            "off": {},
        }
        self.set_set_config(set_config)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        if len(args) != 4:
            return "Usage: define gree_scan fhempy gree_climate scan/name"
        await fhem.readingsSingleUpdate(self.hash, "state", "offline", 1)

        self.name = args[3]
        if self.name == "scan":
            self.create_async_task(self.scan_devices())
            self.set_set_config({})
        else:
            self.create_async_task(self.connect_device())

    async def scan_devices(self, name=None):
        discovery = Discovery()
        device_infos = await discovery.scan(wait_for=5)
        for device_info in device_infos:
            try:
                device = Device(device_info)
                await device.bind()
            except Exception:
                self.logger.exception("Failed to connect to device")
                continue

            if name is None:
                await fhem.CommandDefine(
                    self.hash,
                    (
                        f"gree_{device.device_info.name} "
                        f"fhempy gree_climate {device.device_info.name}"
                    ),
                )
            else:
                if device.device_info.name == name:
                    return device
        return None

    async def connect_device(self):
        self.device = await self.scan_devices(self.name)
        if self.device is None:
            self.logger.error("Couldn't find device")
            return

        await self.update_loop()

    async def update_once(self):
        try:
            await self.device.update_state()
            await self.update_readings()
        except Exception:
            self.logger.exception("Failed to update readings")
            await self.connect_device()

    async def update_loop(self):
        while True:
            if self.device is None:
                # try to connect to device every 10s
                await asyncio.sleep(10)
                await self.connect_device()
                continue
            await self.update_once()
            await asyncio.sleep(self._attr_interval)

    async def update_readings(self):
        await fhem.readingsBeginUpdate(self.hash)
        try:
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "desiredTemp", self.device.target_temperature
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "temperature", self.device.current_temperature
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "mode", Mode(self.device.mode).name
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash,
                "temperature_units",
                TemperatureUnits(self.device.temperature_units).name,
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "fan_speed", FanSpeed(self.device.fan_speed).name
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "fresh_air", "on" if self.device.fresh_air else "off"
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "xfan", "on" if self.device.xfan else "off"
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "anion", "on" if self.device.anion else "off"
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "sleep", "on" if self.device.sleep else "off"
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "light", "on" if self.device.light else "off"
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash,
                "horizontal_swing",
                HorizontalSwing(self.device.horizontal_swing).name,
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash,
                "vertical_swing",
                VerticalSwing(self.device.vertical_swing).name,
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "quiet", "on" if self.device.quiet else "off"
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "turbo", "on" if self.device.turbo else "off"
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "steady_heat", "on" if self.device.steady_heat else "off"
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "power_save", "on" if self.device.power_save else "off"
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "state", "on" if self.device.power else "off"
            )
            # device info
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "ip", self.device.device_info.ip
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "port", self.device.device_info.port
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "brand", self.device.device_info.brand
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "model", self.device.device_info.model
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "name", self.device.device_info.name
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "version", self.device.device_info.version
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "mac", self.device.device_info.mac
            )
        except Exception:
            self.logger.exception("Failed to update_readings")
        await fhem.readingsEndUpdate(self.hash, 1)

    async def set_on(self, hash, params):
        self.device.power = True
        self.create_async_task(self.send_command())

    async def set_off(self, hash, params):
        self.device.power = False
        self.create_async_task(self.send_command())

    async def set_mode(self, hash, params):
        self.device.power = Mode[params["mode"]]
        self.create_async_task(self.send_command())

    async def set_desiredTemp(self, hash, params):
        self.device.target_temperature = params["temperature"]
        self.create_async_task(self.send_command())

    async def set_fan_speed(self, hash, params):
        self.device.fan_speed = FanSpeed[params["speed"]]
        self.create_async_task(self.send_command())

    async def set_swing_horizontal(self, hash, params):
        self.device.horizontal_swing = HorizontalSwing[params["mode"]]
        self.create_async_task(self.send_command())

    async def set_swing_vertical(self, hash, params):
        self.device.vertical_swing = VerticalSwing[params["mode"]]
        self.create_async_task(self.send_command())

    async def set_fresh_air(self, hash, params):
        self.device.fresh_air = True if params["onoff"] == "on" else False
        self.create_async_task(self.send_command())

    async def set_xfan(self, hash, params):
        self.device.xfan = True if params["onoff"] == "on" else False
        self.create_async_task(self.send_command())

    async def set_anion(self, hash, params):
        self.device.anion = True if params["onoff"] == "on" else False
        self.create_async_task(self.send_command())

    async def set_sleep(self, hash, params):
        self.device.sleep = True if params["onoff"] == "on" else False
        self.create_async_task(self.send_command())

    async def set_light(self, hash, params):
        self.device.light = True if params["onoff"] == "on" else False
        self.create_async_task(self.send_command())

    async def set_quiet(self, hash, params):
        self.device.quiet = True if params["onoff"] == "on" else False
        self.create_async_task(self.send_command())

    async def set_turbo(self, hash, params):
        self.device.turbo = True if params["onoff"] == "on" else False
        self.create_async_task(self.send_command())

    async def set_steady_heat(self, hash, params):
        self.device.steady_heat = True if params["onoff"] == "on" else False
        self.create_async_task(self.send_command())

    async def set_power_save(self, hash, params):
        self.device.power_save = True if params["onoff"] == "on" else False
        self.create_async_task(self.send_command())

    async def send_command(self):
        await self.device.push_state_update()
        await self.update_once()



CONF_DEVICES = "devices"
CONF_NAME = "name"
CONF_MAC = "mac"
CONF_DISCOVERY = "discovery"

# Configuration options
CONF_BT_AUTO_RESTART = "bt_auto_restart"
CONF_DECIMALS = "decimals"
CONF_PERIOD = "period"
CONF_LOG_SPIKES = "log_spikes"
CONF_USE_MEDIAN = "use_median"
CONF_ACTIVE_SCAN = "active_scan"
CONF_HCI_INTERFACE = "hci_interface"
CONF_BT_INTERFACE = "bt_interface"
CONF_BATT_ENTITIES = "batt_entities"
CONF_REPORT_UNKNOWN = "report_unknown"
CONF_RESTORE_STATE = "restore_state"
CONF_DEVICE_ENCRYPTION_KEY = "encryption_key"
CONF_DEVICE_DECIMALS = "decimals"
CONF_DEVICE_USE_MEDIAN = "use_median"
CONF_DEVICE_RESTORE_STATE = "restore_state"
CONF_DEVICE_RESET_TIMER = "reset_timer"
CONF_DEVICE_TRACK = "track_device"
CONF_DEVICE_TRACKER_SCAN_INTERVAL = "tracker_scan_interval"
CONF_DEVICE_TRACKER_CONSIDER_HOME = "consider_home"
CONF_DEVICE_DELETE_DEVICE = "delete device"
CONF_PACKET = "packet"
CONF_GATEWAY_ID = "gateway_id"
CONF_UUID = "uuid"

# Default values for configuration options
DEFAULT_BT_AUTO_RESTART = True
DEFAULT_DECIMALS = 1
DEFAULT_PERIOD = 60
DEFAULT_LOG_SPIKES = False
DEFAULT_USE_MEDIAN = False
DEFAULT_ACTIVE_SCAN = False
DEFAULT_BATT_ENTITIES = True
DEFAULT_REPORT_UNKNOWN = "Off"
DEFAULT_DISCOVERY = False
DEFAULT_RESTORE_STATE = False
DEFAULT_DEVICE_MAC = ""
DEFAULT_DEVICE_UUID = ""
DEFAULT_DEVICE_ENCRYPTION_KEY = None
DEFAULT_DEVICE_DECIMALS = "default"
DEFAULT_DEVICE_USE_MEDIAN = "default"
DEFAULT_DEVICE_RESTORE_STATE = "default"
DEFAULT_DEVICE_RESET_TIMER = 35
DEFAULT_DEVICE_TRACKER_SCAN_INTERVAL = 20
DEFAULT_DEVICE_TRACKER_CONSIDER_HOME = 180
DEFAULT_DEVICE_TRACK = False
DEFAULT_DEVICE_DELETE_DEVICE = False

# regex constants for configuration schema
MAC_REGEX = "(?i)^(?:[0-9A-F]{2}[:]){5}(?:[0-9A-F]{2})$"
# MiBeacon V2/V3 uses 24 character long key
AES128KEY24_REGEX = "(?i)^[A-F0-9]{24}$"
# MiBeacon V4/V5 uses 32 character long key
AES128KEY32_REGEX = "(?i)^[A-F0-9]{32}$"

# Dictionary with supported sensors
# Format {device: [averaging sensor list], [instantly updating sensor list],[binary sensor list]}:
# - [averaging sensor list]:            sensors that update the state after avering of the data
# - [instantly updating sensor list]:   sensors that update the state instantly after new data
# - [binary sensor list]:               binary sensors
MEASUREMENT_DICT = {
    "LYWSDCGQ": [["temperature", "humidity", "battery", "rssi"], [], []],
    "LYWSD02": [["temperature", "humidity", "battery", "rssi"], [], []],
    "LYWSD03MMC": [["temperature", "humidity", "battery", "voltage", "rssi"], [], []],
    "XMWSDJ04MMC": [["temperature", "humidity", "battery", "rssi"], [], []],
    "XMMF01JQD": [["rssi"], ["button"], []],
    "HHCCJCY01": [
        ["temperature", "moisture", "conductivity", "illuminance", "battery", "rssi"],
        [],
        [],
    ],
    "GCLS002": [
        ["temperature", "moisture", "conductivity", "illuminance", "rssi"],
        [],
        [],
    ],
    "HHCCPOT002": [["moisture", "conductivity", "rssi"], [], []],
    "WX08ZM": [["consumable", "battery", "rssi"], [], ["switch"]],
    "MCCGQ02HL": [["battery", "rssi"], [], ["opening", "light"]],
    "YM-K1501": [["rssi"], ["temperature"], ["switch"]],
    "YM-K1501EU": [["rssi"], ["temperature"], ["switch"]],
    "V-SK152": [["rssi"], ["temperature"], ["switch"]],
    "SJWS01LM": [["battery", "rssi"], ["button"], ["moisture"]],
    "MJYD02YL": [["battery", "rssi"], [], ["light", "motion"]],
    "MUE4094RT": [["rssi"], [], ["motion"]],
    "RTCGQ02LM": [["battery", "rssi"], ["button"], ["light", "motion"]],
    "MMC-T201-1": [["temperature", "battery", "rssi"], [], []],
    "M1S-T500": [["battery", "rssi"], [], ["toothbrush"]],
    "ZNMS16LM": [["battery", "rssi"], [], ["lock", "fingerprint"]],
    "ZNMS17LM": [["battery", "rssi"], [], ["lock", "fingerprint"]],
    "MJZNMSQ01YD": [["battery", "rssi"], [], ["lock", "fingerprint"]],
    "XMZNMST02YD": [["battery", "rssi"], [], ["lock", "fingerprint"]],
    "CGC1": [["temperature", "humidity", "battery", "rssi"], [], []],
    "CGD1": [["temperature", "humidity", "battery", "rssi"], [], []],
    "CGDK2": [["temperature", "humidity", "battery", "rssi"], [], []],
    "CGG1": [["temperature", "humidity", "battery", "voltage", "rssi"], [], []],
    "CGG1-ENCRYPTED": [
        ["temperature", "humidity", "battery", "voltage", "rssi"],
        [],
        [],
    ],
    "CGH1": [["battery", "rssi"], [], ["opening"]],
    "CGP1W": [["temperature", "humidity", "battery", "pressure", "rssi"], [], []],
    "CGPR1": [["illuminance", "battery", "rssi"], [], ["light", "motion"]],
    "CGDN1": [["temperature", "humidity", "co2", "pm2.5", "pm10", "rssi"], [], []],
    "MHO-C401": [["temperature", "humidity", "battery", "voltage", "rssi"], [], []],
    "MHO-C303": [["temperature", "humidity", "battery", "rssi"], [], []],
    "JQJCY01YM": [
        ["temperature", "humidity", "battery", "formaldehyde", "rssi"],
        [],
        [],
    ],
    "JTYJGD03MI": [["rssi"], ["button", "battery"], ["smoke detector"]],
    "K9B-1BTN": [["rssi"], ["one btn switch"], []],
    "K9B-2BTN": [["rssi"], ["two btn switch left", "two btn switch right"], []],
    "K9B-3BTN": [
        ["rssi"],
        ["three btn switch left", "three btn switch middle", "three btn switch right"],
        [],
    ],
    "YLAI003": [["rssi", "battery"], ["button"], []],
    "YLYK01YL": [["rssi"], ["remote"], ["remote single press", "remote long press"]],
    "YLYK01YL-FANCL": [["rssi"], ["fan remote"], []],
    "YLYK01YL-VENFAN": [["rssi"], ["ventilator fan remote"], []],
    "YLYB01YL-BHFRC": [["rssi"], ["bathroom heater remote"], []],
    "YLKG07YL/YLKG08YL": [["rssi"], ["dimmer"], []],
    "ATC": [
        ["temperature", "humidity", "battery", "voltage", "rssi"],
        [],
        ["switch", "opening"],
    ],
    "Mi Scale V1": [["rssi"], ["weight", "non-stabilized weight"], ["weight removed"]],
    "Mi Scale V2": [
        ["rssi"],
        ["weight", "non-stabilized weight", "impedance"],
        ["weight removed"],
    ],
    "TZC4": [["rssi"], ["weight", "non-stabilized weight", "impedance"], []],
    "Kegtron KT-100": [["rssi"], ["volume dispensed port 1"], []],
    "Kegtron KT-200": [
        ["rssi"],
        ["volume dispensed port 1", "volume dispensed port 2"],
        [],
    ],
    "Smart hygrometer": [
        ["temperature", "humidity", "battery", "voltage", "rssi"],
        [],
        [],
    ],
    "Lanyard/mini hygrometer": [
        ["temperature", "humidity", "battery", "voltage", "rssi"],
        [],
        [],
    ],
    "T201": [["temperature", "humidity", "battery", "voltage", "rssi"], [], []],
    "H5072/H5075": [["temperature", "humidity", "battery", "rssi"], [], []],
    "H5101/H5102/H5177": [["temperature", "humidity", "battery", "rssi"], [], []],
    "H5051": [["temperature", "humidity", "battery", "rssi"], [], []],
    "H5074": [["temperature", "humidity", "battery", "rssi"], [], []],
    "H5178": [
        [
            "temperature",
            "temperature outdoor",
            "humidity",
            "humidity outdoor",
            "battery",
            "rssi",
        ],
        [],
        [],
    ],
    "H5179": [["temperature", "humidity", "battery", "rssi"], [], []],
    "H5183": [["temperature probe 1", "temperature alarm", "rssi"], [], []],
    "Ruuvitag": [
        ["temperature", "humidity", "pressure", "battery", "voltage", "rssi"],
        ["acceleration"],
        ["motion"],
    ],
    "iNode Energy Meter": [["battery", "voltage", "rssi"], ["energy", "power"], []],
    "iNode Care Sensor 1": [
        ["temperature", "battery", "voltage", "rssi"],
        ["acceleration"],
        ["motion"],
    ],
    "iNode Care Sensor 2": [
        ["temperature", "battery", "voltage", "rssi"],
        ["acceleration"],
        ["motion"],
    ],
    "iNode Care Sensor 3": [
        ["temperature", "humidity", "battery", "voltage", "rssi"],
        ["acceleration"],
        ["motion"],
    ],
    "iNode Care Sensor 4": [
        ["temperature", "battery", "voltage", "rssi"],
        ["acceleration"],
        ["motion"],
    ],
    "iNode Care Sensor 5": [
        ["temperature", "battery", "voltage", "rssi"],
        ["acceleration", "magnetic field", "magnetic field direction"],
        ["motion"],
    ],
    "iNode Care Sensor 6": [
        ["temperature", "battery", "voltage", "rssi"],
        ["acceleration"],
        ["motion"],
    ],
    "iNode Care Sensor T": [["temperature", "battery", "voltage", "rssi"], [], []],
    "iNode Care Sensor HT": [
        ["temperature", "humidity", "battery", "voltage", "rssi"],
        [],
        [],
    ],
    "iNode Care Sensor PT": [
        ["temperature", "pressure", "battery", "voltage", "rssi"],
        [],
        [],
    ],
    "iNode Care Sensor PHT": [
        ["temperature", "humidity", "pressure", "battery", "voltage", "rssi"],
        [],
        [],
    ],
    "Blue Puck T": [["temperature", "rssi"], [], []],
    "Blue Coin T": [["temperature", "rssi"], [], []],
    "Blue Puck RHT": [["temperature", "humidity", "rssi"], [], []],
    "HTP.xw": [["temperature", "humidity", "pressure", "rssi"], [], []],
    "HT.w": [["temperature", "humidity", "pressure", "rssi"], [], []],
    "Moat S2": [["temperature", "humidity", "battery", "rssi"], [], []],
    "Tempo Disc THD": [
        ["temperature", "humidity", "dewpoint", "battery", "rssi"],
        [],
        [],
    ],
    "Tempo Disc THPD": [
        ["temperature", "humidity", "pressure", "battery", "rssi"],
        [],
        [],
    ],
    "b-parasite V1.0.0": [
        ["temperature", "humidity", "moisture", "voltage", "rssi"],
        [],
        [],
    ],
    "b-parasite V1.1.0": [
        ["temperature", "humidity", "moisture", "voltage", "rssi", "illuminance"],
        [],
        [],
    ],
    "SmartSeries 7000": [["rssi"], [], ["toothbrush"]],
    "iBBQ-1": [["temperature probe 1", "rssi"], [], []],
    "iBBQ-2": [["temperature probe 1", "temperature probe 2", "rssi"], [], []],
    "iBBQ-4": [
        [
            "temperature probe 1",
            "temperature probe 2",
            "temperature probe 3",
            "temperature probe 4",
            "rssi",
        ],
        [],
        [],
    ],
    "iBBQ-6": [
        [
            "temperature probe 1",
            "temperature probe 2",
            "temperature probe 3",
            "temperature probe 4",
            "temperature probe 5",
            "temperature probe 6",
            "rssi",
        ],
        [],
        [],
    ],
    "IBS-TH": [["temperature", "humidity", "battery", "rssi"], [], []],
    "BEC07-5": [["temperature", "humidity", "rssi"], [], []],
    "iBeacon": [
        ["rssi", "measured power", "cypress temperature", "cypress humidity"],
        ["uuid", "mac", "major", "minor"],
        [],
    ],  # mac can be dynamic
    "AltBeacon": [
        ["rssi", "measured power"],
        ["uuid", "mac", "major", "minor"],
        [],
    ],  # mac can be dynamic
    "MyCO2": [["temperature", "humidity", "co2", "rssi"], [], []],
    "HA BLE DIY": [["temperature", "rssi"], [], []],
}


# Sensor manufacturer dictionary
MANUFACTURER_DICT = {
    "LYWSDCGQ": "Xiaomi",
    "LYWSD02": "Xiaomi",
    "LYWSD03MMC": "Xiaomi",
    "XMWSDJ04MMC": "Xiaomi",
    "XMMF01JQD": "Xiaomi",
    "HHCCJCY01": "Xiaomi",
    "GCLS002": "Xiaomi",
    "HHCCPOT002": "Xiaomi",
    "WX08ZM": "Xiaomi",
    "MCCGQ02HL": "Xiaomi",
    "YM-K1501": "Xiaomi",
    "YM-K1501EU": "Xiaomi",
    "V-SK152": "Viomi",
    "SJWS01LM": "Xiaomi",
    "MJYD02YL": "Xiaomi",
    "MUE4094RT": "Xiaomi",
    "RTCGQ02LM": "Xiaomi",
    "MMC-T201-1": "Xiaomi",
    "M1S-T500": "Xiaomi Soocas",
    "ZNMS16LM": "Xiaomi Aqara",
    "ZNMS17LM": "Xiaomi Aqara",
    "MJZNMSQ01YD": "Xiaomi",
    "XMZNMST02YD": "Xiaomi",
    "CGC1": "Qingping",
    "CGD1": "Qingping",
    "CGDK2": "Qingping",
    "CGG1": "Qingping",
    "CGG1-ENCRYPTED": "Qingping",
    "CGH1": "Qingping",
    "CGP1W": "Qingping",
    "CGPR1": "Qingping",
    "CGDN1": "Qingping",
    "MHO-C401": "Miaomiaoce",
    "MHO-C303": "Miaomiaoce",
    "JQJCY01YM": "Honeywell",
    "JTYJGD03MI": "Honeywell",
    "YLAI003": "Yeelight",
    "YLYK01YL": "Yeelight",
    "YLYK01YL-FANCL": "Yeelight",
    "YLYK01YL-VENFAN": "Yeelight",
    "YLYB01YL-BHFRC": "Yeelight",
    "YLKG07YL/YLKG08YL": "Yeelight",
    "K9B-1BTN": "Linptech",
    "K9B-2BTN": "Linptech",
    "K9B-3BTN": "Linptech",
    "ATC": "ATC",
    "Mi Scale V1": "Xiaomi",
    "Mi Scale V2": "Xiaomi",
    "TZC4": "Xiaogui",
    "Kegtron KT-100": "Kegtron",
    "Kegtron KT-200": "Kegtron",
    "Smart hygrometer": "Thermoplus",
    "Lanyard/mini hygrometer": "Thermoplus",
    "T201": "Brifit",
    "H5072/H5075": "Govee",
    "H5101/H5102/H5177": "Govee",
    "H5051": "Govee",
    "H5074": "Govee",
    "H5178": "Govee",
    "H5179": "Govee",
    "H5183": "Govee",
    "Ruuvitag": "Ruuvitag",
    "iNode Energy Meter": "iNode",
    "iNode Care Sensor 1": "iNode",
    "iNode Care Sensor 2": "iNode",
    "iNode Care Sensor 3": "iNode",
    "iNode Care Sensor 4": "iNode",
    "iNode Care Sensor 5": "iNode",
    "iNode Care Sensor 6": "iNode",
    "iNode Care Sensor T": "iNode",
    "iNode Care Sensor HT": "iNode",
    "iNode Care Sensor PT": "iNode",
    "iNode Care Sensor PHT": "iNode",
    "Blue Puck T": "Teltonika",
    "Blue Coin T": "Teltonika",
    "Blue Puck RHT": "Teltonika",
    "HTP.xw": "SensorPush",
    "HT.w": "SensorPush",
    "MyCO2": "Sensirion",
    "Moat S2": "Moat",
    "Tempo Disc THD": "BlueMaestro",
    "Tempo Disc THPD": "BlueMaestro",
    "b-parasite V1.0.0": "rbaron",
    "b-parasite V1.1.0": "rbaron",
    "SmartSeries 7000": "Oral-B",
    "iBBQ-1": "Inkbird",
    "iBBQ-2": "Inkbird",
    "iBBQ-4": "Inkbird",
    "iBBQ-6": "Inkbird",
    "IBS-TH": "Inkbird",
    "BEC07-5": "Jinou",
    "iBeacon": "Apple",
    "AltBeacon": "Radius Networks",
    "HA BLE DIY": "Home Assistant DIY",
}

# Renamed model dictionary
RENAMED_MODEL_DICT = {
    "H5051/H5074": "H5074",
    "IBS-TH2": "IBS-TH",
}

# Selection list for report_uknown
REPORT_UNKNOWN_LIST = [
    "ATC",
    "BlueMaestro",
    "Brifit",
    "Govee",
    "HA BLE",
    "Inkbird",
    "iNode",
    "iBeacon",
    "Jinou" "Kegtron",
    "Mi Scale",
    "Moat",
    "Oral-B",
    "Qingping",
    "rbaron",
    "Ruuvitag",
    "Sensirion",
    "SensorPush",
    "Teltonika",
    "Thermoplus",
    "Xiaogui",
    "Xiaomi",
    "Other",
    "Off",
    False,
]



import asyncio
import time

from fhempy.lib.fusionsolar.fusionsolar_api import FusionSolarRestApi

from .. import fhem, generic


class fusionsolar(generic.FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self._stationname = None
        self._sessionid = None

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        if len(argsh) > 0 and argsh[list(argsh)[0]][:4] == "http":
            await fhem.readingsSingleUpdate(
                hash, "state", "Please read HELP to change define", 1
            )
            return

        if not (len(args) == 5 or len(args) == 6):
            return (
                "Usage: define my_solar fhempy fusionsolar "
                "USERNAME PASSWORD [REGION]"
            )

        self._username = args[3]
        self._password = args[4]
        self._region = args[5] if len(args) == 6 else "region01eu5"

        await fhem.readingsSingleUpdate(hash, "state", "connecting", 1)
        self.create_async_task(self.update())

    async def update(self):
        self.restapi = FusionSolarRestApi(
            self.logger,
            self._username,
            self._password,
            self._region,
        )
        if await self.restapi.login():
            await self.update_readings()
        else:
            await fhem.readingsSingleUpdate(self.hash, "state", "login failed", 1)

    async def update_readings(self):
        while True:
            await fhem.readingsBeginUpdate(self.hash)
            try:
                await self.restapi.update()

                await fhem.readingsBulkUpdate(
                    self.hash, "from_grid_power", self.restapi.from_grid_power
                )
                await fhem.readingsBulkUpdate(
                    self.hash, "to_grid_power", self.restapi.to_grid_power
                )
                await fhem.readingsBulkUpdate(
                    self.hash, "electrical_load", self.restapi.electrical_load
                )
                await fhem.readingsBulkUpdate(
                    self.hash, "grid_power", self.restapi.grid_power
                )
                await fhem.readingsBulkUpdate(
                    self.hash,
                    "inverter_output_power",
                    self.restapi.inverter_output_power,
                )
                await fhem.readingsBulkUpdate(
                    self.hash,
                    "string_output_power",
                    self.restapi.string_output_power,
                )
                await fhem.readingsBulkUpdate(
                    self.hash,
                    "daily_self_use_ratio",
                    self.restapi.daily_self_use_ratio,
                )
                await fhem.readingsBulkUpdate(
                    self.hash,
                    "station",
                    self.restapi.station,
                )
                await fhem.readingsBulkUpdate(
                    self.hash,
                    "co2_saved",
                    self.restapi.co2_saved,
                )
                await fhem.readingsBulkUpdate(
                    self.hash,
                    "daily_self_use_energy",
                    self.restapi.daily_self_use_energy,
                )
                await fhem.readingsBulkUpdate(
                    self.hash,
                    "daily_self_use_solar_ratio",
                    self.restapi.daily_self_use_solar_ratio,
                )
                await fhem.readingsBulkUpdate(
                    self.hash,
                    "daily_use_energy",
                    self.restapi.daily_use_energy,
                )
                await fhem.readingsBulkUpdate(
                    self.hash,
                    "grid_connected_time",
                    self.restapi.grid_connected_time,
                )
                await fhem.readingsBulkUpdate(
                    self.hash,
                    "installed_capacity",
                    self.restapi.installed_capacity,
                )
                await fhem.readingsBulkUpdate(
                    self.hash,
                    "total_current_day_energy",
                    self.restapi.total_current_day_energy,
                )
                await fhem.readingsBulkUpdate(
                    self.hash,
                    "total_current_month_energy",
                    self.restapi.total_current_month_energy,
                )
                await fhem.readingsBulkUpdate(
                    self.hash,
                    "total_current_year_energy",
                    self.restapi.total_current_year_energy,
                )
                await fhem.readingsBulkUpdate(
                    self.hash,
                    "total_lifetime_energy",
                    self.restapi.total_lifetime_energy,
                )

                # battery values
                if self.restapi.battery_soc is not None:
                    await fhem.readingsBulkUpdate(
                        self.hash,
                        "battery_soc",
                        self.restapi.battery_soc,
                    )
                    await fhem.readingsBulkUpdate(
                        self.hash,
                        "battery_power",
                        self.restapi.battery_power,
                    )
                    await fhem.readingsBulkUpdate(
                        self.hash,
                        "battery_charge_capacity",
                        self.restapi.battery_charge_capacity,
                    )
                    await fhem.readingsBulkUpdate(
                        self.hash,
                        "battery_discharge_capacity",
                        self.restapi.battery_discharge_capacity,
                    )

                # string details
                for string in self.restapi.string_details:
                    string_detail = self.restapi.string_details[string]
                    await fhem.readingsBulkUpdate(
                        self.hash, f"{string}_voltage", string_detail["voltage"]
                    )
                    await fhem.readingsBulkUpdate(
                        self.hash, f"{string}_current", string_detail["current"]
                    )

                await fhem.readingsBulkUpdateIfChanged(self.hash, "state", "connected")
            except Exception:
                await fhem.readingsBulkUpdateIfChanged(self.hash, "state", "failed")
                self.logger.exception("Failed to update readings")
            await fhem.readingsEndUpdate(self.hash, 1)
            # calculate next full 5min+30s
            next_sec = (
                (4 - time.localtime().tm_min % 5) * 60
                + 60
                - time.localtime().tm_sec
                + 30
            )
            if next_sec > 180:
                next_sec = 180
            await asyncio.sleep(next_sec)



"""
Starts a service to scan in intervals for new devices.
"""

from fhempy.lib.core.ssdp import ssdp
from fhempy.lib.generic import FhemModule

from .. import fhem


class discover_upnp(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self.set_set_config({})
        self.hash = None
        self.create_devs = {}

    async def found_device(self, upnp_device):
        if upnp_device and upnp_device.udn:
            await fhem.readingsSingleUpdate(
                self.hash,
                upnp_device.udn.replace(":", ".")
                + "_"
                + upnp_device.device_type.replace(":", "."),
                upnp_device.friendly_name,
                0,
            )
            if upnp_device.device_type == "urn:schemas-upnp-org:device:MediaRenderer:1":
                self.create_devs[upnp_device.udn] = {
                    "name": "".join(filter(str.isalnum, upnp_device.friendly_name))
                    + "_"
                    + upnp_device.device_type.split(":")[-2],
                    "devname": "".join(filter(str.isalnum, upnp_device.friendly_name))
                    + "_"
                    + upnp_device.device_type.split(":")[-2],
                }
                set_config = {"create": {"args": ["device"], "options": ""}}
                set_devs = []
                for dev in self.create_devs:
                    set_devs.append(self.create_devs[dev]["name"])
                set_config["create"]["options"] = ",".join(set_devs)
                self.set_set_config(set_config)

    async def removed_device(self, upnp_device):
        return

    # FHEM Define
    async def Define(self, hash, args, argsh):
        """Start a discovery service."""
        await super().Define(hash, args, argsh)
        ssdp.getInstance(self.logger).register_listener(self)
        await ssdp.getInstance(self.logger).start_search()
        await fhem.readingsSingleUpdate(hash, "state", "active", 0)

        if await fhem.AttrVal(self.hash["NAME"], "icon", "") == "":
            await fhem.CommandAttr(self.hash, self.hash["NAME"] + " icon rc_SEARCH")

    async def set_create(self, hash, params):
        for udn in self.create_devs:
            if self.create_devs[udn]["name"] == params["device"]:
                await fhem.CommandDefine(
                    self.hash,
                    self.create_devs[udn]["devname"] + " fhempy dlna_dmr " + udn,
                )

    # FHEM Undefine
    async def Undefine(self, hash):
        await ssdp.getInstance(self.logger).stop_search()
        await super().Undefine(hash)
